const fs = require('fs');

const Report = require('../lib/report.js');

/*
 * The various sanitizers are currently vehemently against making
 * machine-readable output. We have to apply various heuristics to
 * make this data useful.
 */
function tsan_callstack_parse(line) {
  var f = {
    count: false,
    pc: false,
    function: null,
  };
  var frame = {};

  /* First, attempt to extract the object. */
  var i_o = line.lastIndexOf(')');
  var object_end = i_o;

  /* Now, extract the outermost matching "(" character. */
  var c = [0, 1];
  while (i_o-- > 0) {
    if (line[i_o] === ')') {
      c[1]++;
    } else if (line[i_o] === '(') {
      c[0]++;
    }

    /* We maybe extract the object. */
    if (c[0] === c[1]) break;
  }

  if (i_o === 0) return null;

  var object = line.substring(i_o + 1, object_end);
  var object_name = object.split('+')[0];

  /* We have extracted the object name. */
  frame.object = object_name;

  /* Fake the program counter. */
  frame.registers = {};
  frame.registers.pc = object;

  /*
   * The grammar is ambiguous, we will assume that the second space
   * is the function name.
   */
  var tokens = line.split(' ');
  for (var i = 0; i < tokens.length; i++) {
    if (tokens[i] !== '') break;
  }

  if (tokens[i + 1] && tokens[i + 1] !== '<null>') {
    var fn = tokens[i + 1];

    /* Ignore any function arguments. */
    fn = fn.split('(')[0];
    frame['function'] = tokens[i + 1];
  }

  /* Now, we will try to at least derive a source file name. */
  for (var i = 0; i < tokens.length; i++) {
    var l_i = tokens[i].lastIndexOf(':');

    if (l_i > 0) {
      /* This is a line number probably. */
      frame['line'] = tokens[i].substring(l_i + 1, tokens[i].length);

      /* At least attempt to extract the basename of the file. */
      var file = tokens[i].substring(0, l_i);

      var l_s;

      l_s = file.lastIndexOf('/');
      if (l_s === -1) l_s = file.lastIndexOf('\\');

      if (l_s >= 0) {
        file = file.substring(l_s + 1, file.length);
      }

      frame['file'] = file;
    }
  }

  return frame;
}

function callstack_parse(line) {
  var tokens = line.split(' ');

  var f = {
    count: false,
    pc: false,
    function: null,
  };
  var frame = {};

  // Trims leading whitespace
  for (var i = 0; i < tokens.length; i++) {
    if (tokens[i] === '') continue;

    break;
  }

  /*
   * The first two columns are an offset, address and the last column
   * is a location. Anything after in is a symbol.
   */
  if (tokens[i] && tokens[i][0] === '#') {
    f.count = true;
    i++;
  } else {
    console.error('Malformed line: ' + line);
    return null;
  }

  var address = tokens[i];

  if (address[0] !== '0' || address[1] !== 'x') {
    console.error('Invalid memory address: ' + tokens[i]);
    return null;
  }

  frame.registers = {};
  frame.registers.pc = address;
  i++;

  /* Fast track to the "in". */
  var i_o = i;

  while (tokens[i] != 'in' && i < tokens.length) i++;

  /* It is possible all this data is missing. */
  if (i === tokens.length) {
    i = i_o;
  }

  /*
   * Everything in between is the function name, probably, assuming
   * no one changes ASAN output.
   */
  var name = '';
  for (i++; i < tokens.length - 1; i++) name += tokens[i] + ' ';
  name = name.substring(0, name.length - 1);

  frame['function'] = name;

  /* Go to the last token. */
  var ln = tokens[tokens.length - 1];

  /* Attempt to disambiguate object versus source file. */
  if (ln[0] === '(') {
    frame.object = ln.split('+')[0].replace('(', '');
  } else if (ln[0] !== '(' && ln[ln.length - 1] === ')') {
    /* This is an unknown module, we prefer to ignore this frame. */
    return null;
  } else {
    /* Pray to god that this is a source file location. */
    var split = ln.lastIndexOf(':');

    if (split >= 0) {
      var f = ln.substring(0, split);
      var l = ln.substring(split + 1, ln.length);

      frame['file'] = f;
      frame['line'] = parseInt(l);
    } else {
      frame['function'] = ln;
    }
  }

  return frame;
}

class sanitizers {
  constructor(options) {
    this.options = options;
    this.file = options.file;
    this.data = options.data;
    this.exclude = options.exclude;
    if (!this.exclude) this.exclude = [];
  }

  extractLsan(report, lines, cause, pid, raw, cb) {
    let type_set = null;

    /*
     * Boundaries are located at "X leak of N byte in Y object",
     * the subsequent line is a callstack.
     */
    report
      .set('attribute', 'tool', 'LeakSanitizer')
      .add('classifiers', 'LeakSanitizer');

    for (var i = 0; i < lines.length; i++) {
      var words = lines[i].split(' ');

      if (
        words[1] !== 'leak' ||
        words[2] !== 'of' ||
        words[4] !== 'byte(s)' ||
        words[5] !== 'in' ||
        words[7] !== 'object(s)'
      ) {
        /* console.log('Ignored line: ' + lines[i]); */
        continue;
      }

      if (type_set === null && lines[i].indexOf('Indirect leak of') >= 0) {
        type_set = 'IndirectLeak';
      } else if (type_set === null && lines[i].indexOf('Direct leak of') >= 0) {
        type_set = 'DirectLeak';
      }

      /* Every leak is a new report. */
      var text = lines[i];

      /* Extract attributes. */
      var leaked_bytes = parseInt(words[3]);
      var leaked_blocks = parseInt(words[6]);

      var faulting_frame;
      var frames = [];

      for (i++; i < lines.length; i++) {
        if (lines[i] === '') break;

        var frame = callstack_parse(lines[i]);
        if (frame !== null) {
          frames.push(frame);
          if (!faulting_frame) faulting_frame = frame;
        }
      }

      if (frames.length > 0) {
        var nr = new Report(report);
        var thread = nr.thread(pid);

        nr.set('attribute', 'leaked.bytes', leaked_bytes + '').set(
          'attribute',
          'leaked.blocks',
          leaked_blocks + '',
        );

        if (type_set !== null) {
          nr.add('classifiers', type_set);
          type_set = null;
        }

        faulting_frame.text = text;
        for (var j = 0; j < frames.length; j++) {
          thread.frame(frames[j]);
        }
      }
    }
  }

  extractAsan(report, lines, cause, pid, raw, cb) {
    var self = this;
    var tokens = cause.split(' ');

    report.set('attribute', 'tool', 'AddressSanitizer');
    report.add('classifiers', 'AddressSanitizer');

    /* Attempt to determine the shape of the error message. */
    if (
      tokens[2] === 'on' &&
      tokens[3] === 'address' &&
      tokens[4].indexOf('0x') === 0 &&
      tokens[5] === 'at' &&
      tokens[6] === 'pc'
    ) {
      var registers = {};
      var thread = report.thread(pid);

      report.set('attribute', 'fault.address', tokens[4]);

      /* Extract registers. */
      for (var i = 6; i < tokens.length; i++) {
        if (tokens[i] === 'pc' || tokens[i] === 'bp' || tokens[i] === 'sp') {
          registers[tokens[i]] = tokens[i + 1];
          i++;
        }
      }

      /* Serialize into the callstack. */
      var text = lines[2];
      var faulting_frame;

      /*
       * Extract the callstack. It is the subsequent line to ' of size ' and
       * 'at' line.
       */
      for (var i = 3; i < lines.length; i++) {
        if (lines[i] === '') break;

        if (
          lines[i].indexOf('by thread') >= 0 ||
          lines[i].indexOf('is located') >= 0 ||
          lines[i].indexOf('by thread T') >= 0
        ) {
          break;
        }

        var frame = callstack_parse(lines[i]);
        if (frame !== null) {
          thread.frame(frame);
          if (!faulting_frame) faulting_frame = frame;
        }
      }

      faulting_frame.text = raw;
      if (Object.keys(registers).length > 0)
        faulting_frame.registers = registers;
    } else {
      var thread = report.thread(pid);

      /* For any other shape, pray that the subsequent line is a callstack. */
      for (var i = 2; i < lines.length; i++) {
        if (lines[i] === '') break;

        var frame = callstack_parse(lines[i]);
        if (frame !== null) {
          thread.frame(frame);
          if (!faulting_frame) faulting_frame = frame;
        }
      }

      faulting_frame.text = raw;
    }

    /* Find the summary lines. */
    for (var i = 0; i < lines.length; i++) {
      if (lines[i].indexOf('SUMMARY: ') !== 0) continue;

      /* Find the summary line, extract a classifier. */
      var tc = lines[i].split(':');
      var classifier = tc[2].split(' ')[1];

      report.add('classifiers', classifier);
    }
  }

  extractError(error, root) {
    /*
     * Every error block begins with:
     * ==<PID>==ERROR: <Tool>: <Description>
     */
    var self = this;
    var lines = error.split('\n');

    if (lines.length < 3) return;

    /* The first line is the split line. */
    var header = lines[1];

    var token = header.split('==');
    var pid = token[1];

    /* Extract the tool. */
    var tokens = header.split(':');
    if (tokens.length <= 1) return;

    var tool = tokens[1].match(/[A-Za-z]+/)[0];

    /* Extract the actual error and registers now. */
    var cause = tokens[2];

    /*
     * Finally, we extract errors according to the type of error.
     * There are two main sections. A callstack used for deduplication,
     * and then the actual body.
     */
    var report = new Report(root);
    report.set('attribute', 'pid', pid);

    if (tool === 'AddressSanitizer' && self.exclude.indexOf('errors') === -1) {
      self.extractAsan(report, lines, cause, pid, error);
    } else if (
      tool === 'LeakSanitizer' &&
      self.exclude.indexOf('leaks') === -1
    ) {
      self.extractLsan(root, lines, cause, pid, error);
    }
  }

  extract(options, errors, raw, cb) {
    var self = this;
    var report = new Report();

    report.checksum(raw);
    if (options && options.attributes) {
      for (var a in options.attributes)
        report.set('attribute', a, options.attributes[a]);
    }

    for (var i = 0; i < errors.length; i++) {
      self.extractError(errors[i], report);
    }

    return cb(null, report);
  }

  extractTsanError(error, report) {
    if (error === '') return;

    var lines = error.split('\n');

    for (var i = 0; i < lines.length; i++) {
      var rp;

      if (lines[i].indexOf('WARNING: ThreadSanitizer: ') === -1) continue;

      /* We are now at an actual error. Try to extract the context. */
      var tokens = lines[i].split(': ');

      rp = new Report(report);
      rp.set('attribute', 'tool', tokens[1]).add('classifiers', tokens[1]);

      var classifier = tokens[2].split(' (');

      classifier[0] = classifier[0].replace(/ /g, '-');
      if (classifier[0] && classifier[0].length > 0)
        rp.add('classifiers', classifier[0]);

      if (classifier[1] && classifier[1].length > 0) {
        /* Attempt to extract PID. */
        var pid = classifier[1].split('=')[1];
        if (pid && pid.length > 0) {
          /* Remove anything that isn't a digit. */
          pid = pid.replace(/\D/g, '');
          if (pid) rp.add('attribute', 'pid', pid);
        }
      }

      /*
       * Now, we will seek to the first callstack and attempt to serialize
       * that.
       */
      for (i++; i < lines.length; i++) {
        if (lines[i].indexOf('#0') >= 0) break;
      }

      /* If no callstack was found, we have failed to find a callstack. */
      if (i === lines.length) return;

      /* Otherwise, we have found a callstack. */
      var thread = rp.thread('Thread');
      var faulting_frame;
      for (; i < lines.length; i++) {
        /* We are probably done with the callstack. */
        if (lines[i].length < 8) break;

        /* Extract the callstack. */
        var frame = tsan_callstack_parse(lines[i]);

        if (frame) {
          if (!faulting_frame) faulting_frame = frame;

          thread.frame(frame);
        }
      }

      /*
       * If a faulting frame is found, attach the raw buffer to it.
       */
      if (faulting_frame) {
        faulting_frame.text = error;
        return;
      }
    }
  }

  extractTsan(options, errors, raw, cb) {
    var self = this;
    var report = new Report();

    report.checksum(raw);
    if (options && options.attributes) {
      for (var a in options.attributes)
        report.set('attribute', a, options.attributes[a]);
    }

    for (var i = 0; i < errors.length; i++) {
      self.extractTsanError(errors[i], report);
    }

    return cb(null, report);
  }

  parseData(options, cb) {
    var self = this;

    /* Check if we are parsing thread sanitizer. */
    if (self.data.indexOf('WARNING: ThreadSanitizer') >= 0) {
      var errors;

      if (self.data.indexOf('==================\n') > 0) {
        errors = self.data.split('==================\n');
      } else {
        errors = self.data.split('WARNING: ThreadSanitizer: ');

        /* Go through and prefix the string. */
        for (var i = 0; i < errors.length; i++) {
          if (errors[i].length <= 8) continue;

          errors[i] = 'WARNING: ThreadSanitizer: ' + errors[i];
        }
      }

      return self.extractTsan(options, errors, self.data, cb);
    }

    var errors = self.data.split(
      '=================================================================',
    );
    return self.extract(options, errors, self.data, cb);
  }

  parse(options, cb) {
    var self = this;

    if (self.data) {
      return self.parseData(options, cb);
    } else if (self.file) {
      fs.readFile(this.file, 'utf8', function (e, d) {
        if (e) {
          return cb(e, null);
        }

        self.data = d;
        return self.parseData(options, cb);
      });
    } else {
      return cb('no data was supplied');
    }
  }
}

module.exports = sanitizers;

//-- vim:ts=2:et:sw=2:st=2
